# Make Spatiotemporal Quantile Regression Data.r
# Generated outcome data from a class of location-scale t distributions (non-Gaussian). 
# The location parameters are systematically determined by additive spatiotemporal effects,
# the scale parameters vary according to a gamma (2, rate = 2) distribution. 
# To emulate the DSR data, we use NC adjacencies and assume a five-year period (2007 - 2011) 
# Infiles:  1) C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/County A.txt (NC Adj Matrix)
#  	        2) C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/Time K.txt (Time Adj Matrix)
#           3) C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo//writeDatafileR.txt
# Outfile:  C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/spatiotemporal quantile regression data.txt
# Mar 22, 2015
#########################
library(mvtnorm)
library(spam)
source("C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo//writeDatafileR.txt")

set.seed(032215)    
n <- 153                                          # No. of BGs
m <- 5                                            # No. of years
nij <- 100                                        # No. of records per BG and year
ni <- m * nij
nj <- n * nij
ids <- rep(rep(1:n, rep(nij, n)), m)              # ID variable for space/BG 
idt <- rep(1:m, rep(nj, m))                       # ID variable for time/year
N <- n * m * nij

##################
# Adjacency Info #
##################
# NC Adjacency Matrix
A <- dget("C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/County A.txt")
mi <- apply(A, 1, sum)                            # No. neighbors
Q <- as.spam(diag(mi)) - as.spam(A) + diag(.0001, n)

# Temporal Adjacency Matrix
K <- dget("C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/Time K.txt")
K <- as.spam(K) + diag(.0001, m) 

# Centering Matrix for Theta
Ccol <- diag(n) - tcrossprod(rep(1, n))/n         # centering by column
Crow <- diag(m) - tcrossprod(rep(1, m))/m         # centering by row

######################
# Space Time Effects #
######################
# Variance parameters
sdphi <- sqrt(2)                                    # spatial sd
sdpsi <- sqrt(2)                                    # temporal sd
sdtheta <- sqrt(4)                                  # interaction sd
covphi <- sdphi^2 * solve(Q)                        # spatial cov matrix
covpsi <- sdpsi^2 * solve(K)                        # temporal cov matrix
covtheta <- sdtheta^2 * solve(K %x% Q)              # interaction cov matrix

#############################
# Random effects simulation #
#############################
phi <- c(rmvnorm(1, sigma = covphi))                # spatial CAR effects
psi <- c(rmvnorm(1, sigma = covpsi))                # temporal CAR effects
theta <- c(rmvnorm(1, sigma = covtheta))            # space-time MCAR effects
phic <- rep(rep(phi - mean(phi), rep(nij, n)), m)   # replicate by time
psic <- rep(psi - mean(psi), rep(nj, m))            # replicate by space
thetam <- Ccol %*% matrix(theta, n, m) %*% Crow
thetac <- rep(c(thetam), each = nij)                # replicate by record

# functions related to location-scale t family
qnst <- function(p, mu, sigma, nu) return(mu + sigma * qt(p, df = nu)) # quantile function for t r.v. with location mu and scale sigma, df = nu
rnst <- function(n, mu, sigma, nu) return(mu + sigma * rt(n, df = nu)) # generate n t r.v. with location mu and scale sigma, df = nu

######################
# Simulation Outcome #
######################
e <- e25 <- e50 <- e75 <- e90 <- NULL               # "observed" residual and true quantile residual
nu <- 2                                             # df
sigma <- rgamma(n * m, 2, 2)                        # scale parameter
for(i in 1:m){
  for(j in 1:n){
    year <- 2006 + i
    BG <- j
    s <- (i - 1)*n + j
    e<-c(e, rnst(n = nij, mu = 0, sigma = sigma[s], nu = nu))                       # 'observed' residual
    e25 <- c(e25, rep(qnst(p = 0.25, mu = 0, sigma = sigma[s], nu = nu), nij))      # true 25th quantile of each t density
    e50 <- c(e50, rep(qnst(p = 0.50, mu = 0, sigma = sigma[s], nu = nu), nij))      # true 50th quantile of each t density
    e75 <- c(e75, rep(qnst(p = 0.75, mu = 0, sigma = sigma[s], nu = nu), nij))      # true 75th quantile of each t density
    e90 <- c(e90, rep(qnst(p = 0.90, mu = 0, sigma = sigma[s], nu = nu), nij))      # true 90th quantile of each t density
  }  
}

eta <- phic + psic + thetac                         # location parameter
y <- eta + e                                        # outcomes
y25 <- eta + e25                                    # 25th quantile
y50 <- eta + e50                                    # 50th quantile
y75 <- eta + e75                                    # 75th quantile
y90 <- eta + e90                                    # 90th quantile

######################
#     Save Data      #
######################
data <- list(ids = ids, idt = idt, y = y, y25 = y25, y50 = y50, y75 = y75, y90 = y90)
writeDatafileR(data, "C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/spatiotemporal quantile regression data.txt")