# Spatiotemporal Quantile Regression.R
# Runs MCMC for fitting the Spatiotemporal Quantile Regression Model with the simulated data
# CONDITIONAL UPDATE FOR LONG VECTOR THETA -- FASTER
# Infile:   1) Data = C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/spatiotemporal quantile regression data.txt
#  	        2) County A.txt = NC County Adjacency Matrix 
#           3) Time K.txt = Temporal Adjacency Matrix
# Oct 22, 2015
##############################
library(mvtnorm)
library(spam)
library(statmod) 
set.seed(032215)    

########
# Data #
########
# Takes a minute to load the 'large' list of data
data <- dget("C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/spatiotemporal quantile regression data.txt")
n <- max(data$ids)                                            # No. of BGs
m <- max(data$idt)                                            # No. of years
nij <- tapply(data$y, list(data$ids, data$idt), length)       # No. of records per BG/year
ni <- apply(nij, 1, sum)                                      # No. of records per BG across years
nj <- apply(nij, 2, sum)                                      # No. of records per year across BGs
N <- sum(nij)                                                 # No. of records in total

nsim <- 1000                                                # length of chain
burn <- 500                                                  # half as burn-in

############################################################################
# Quantile Level -- Change tau between (0, 1) to model different quantiles #
############################################################################
tau <- 0.50                                                   # Median regression

##################
# Adjacency Info #
##################
# NC Adjacency Matrix
A <- dget("C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/County A.txt")
mi <- apply(A, 1, sum)                            # No. neighbors
Q <- as.spam(diag(mi)) - as.spam(A) + diag(.0001, n)

# Temporal Adjacency Matrix
K <- dget("C:/Users/Fan/Desktop/Biostat 220/Publication/STQR_demo/Time K.txt")
K <- as.spam(K) + diag(.0001, m) 

# Centering Matrix for Theta
Ccol <- diag(n) - tcrossprod(rep(1, n))/n         # centering by column
Crow <- diag(m) - tcrossprod(rep(1, m))/m         # centering by row

# Outcome
y <- data$y                                       # outcome records sorted by space and time

##########
# Priors #
##########
m0 <- 0                                           # Prior mean for intercept     
prec0 <- 10^-3                                    # Prior precision for intercept
a <- b <- c <- d <- e <- f <- g <- h <- .001      # Non-informative Gamma shape parameters

#########
# Inits #
#########
tauphi <- taupsi <- tautheta <- delta2 <- 1       # pecision parameters and delta
v <- w <- rep(1, N)                               # augmented latent weights
beta <- 0                                         # intercept 
phi <- rep(0, n)                                  # spatial effects
psi <- rep(0, m)                                  # temporal effects
theta <- rep(0, n * m)                            # interaction effects

####################################
# Special Quantile Specific Values #
# Details in Equaion (10)          #
####################################
xi <- (1 - 2*tau) / (tau * (1-tau))
zeta <- 2 / (tau * (1-tau))
taue <- delta2 / zeta                             # define a pseudo-precision involving delta

#######################################
# # Replicate effects for each record #
#######################################

Phic <- rep(rep(phi, m), c(nij))                  # spatial effects for each record
Psic <- rep(rep(psi, each = n), c(nij))           # temporal effects for each record
thetam <- matrix(theta, n, m)                     # interaction in matrix form
Thetac <- rep(c(thetam), c(nij))                  # interaction effects for each record

#########
# Store #
#########
Beta <- S2phi <- S2psi <- S2theta <- Delta2 <- rep(0, nsim)  
Phis <- matrix(0, nsim, n)
Psis <- matrix(0, nsim, m)
Thetas <- matrix(0, nsim, n*m)

#########
# Gibbs #
#########
ptm<-proc.time()

for(i in 1:nsim){
  
  #########################################
  # Update fixed coefficients (intercept) #
  #########################################
  vbeta <- 1 / (taue * sum(v) + prec0)
  mbeta <- vbeta * (taue * sum(v * (y - xi*w - Phic - Psic - Thetac)))
  beta <- rnorm(1, mbeta, sqrt(vbeta))
  
  ##############################
  # Update Spatial CAR effects #
  ############################## 
  vphi <- taue * diag(tapply(v, ids, FUN = sum)) + tauphi * Q
  mphi <- taue * tapply(v * (y - xi*w - beta - Psic - Thetac), ids, FUN = sum)
  phi <- rmvnorm.canonical(1, mphi, vphi)[1,]
  phi <- phi - mean(phi)
  Phic <- rep(rep(phi, m), c(nij))
  
  ###############################
  # Update Temporal CAR effects #
  ############################### 
  vpsi <- taue * diag(tapply(v, idt, FUN = sum)) + taupsi * K
  mpsi <- taue * tapply(v * (y - xi*w - beta - Phic - Thetac), idt, FUN = sum)
  psi <- rmvnorm.canonical(1, mpsi, vpsi)[1,]
  psi <- psi - mean(psi)
  Psic <- rep(rep(psi, each = n), c(nij))
  
  ######################################################
  # Update Multivariate CAR space-time interaction     #
  # Conditional update by year to speed up the sampler #
  ###################################################### 
  # update BG-year specific quantities involving latent weights
  vsumst <- tapply(v, list(ids, idt), FUN = sum)
  vysumst <- tapply(v * (y - xi*w - beta - Phic - Psic), list(ids, idt), FUN = sum)
  
  # conditional updates
  vtheta1 <- taue * diag(vsumst[,1]) + tautheta * Q 
  mtheta1 <- taue * vysumst[,1] + tautheta * Q %*% thetam[,2]
  thetam[,1] <- rmvnorm.canonical(1, mtheta1, vtheta1)[1,]                               # conditional update theta1 in year 1
  
  for(j in 2:(m-1)){
    vthetaj <- taue * diag(vsumst[,j]) + 2 * tautheta * Q
    mthetaj <- taue * vysumst[,j] + tautheta * Q %*% (thetam[,j-1] + thetam[,j+1])
    thetam[,j] <- rmvnorm.canonical(1, mthetaj, vthetaj)[1,]                             # conditional update thetaj in year j
  }
  
  vthetam <- taue * diag(vsumst[,m]) + tautheta * Q                                      
  mthetam <- taue * vysumst[,m] + tautheta * Q %*% thetam[,m-1]
  thetam[,m] <- rmvnorm.canonical(1, mthetam, vthetam)[1,]                               # conditional update thetam in year m
  thetam <- Ccol %*% thetam %*% Crow                                                     # centering by row and column to ensure identifiability
  Thetac <- rep(c(thetam), c(nij))                                                       # replicate by record
  
  ################################
  # Update Spatial CAR precision #
  ################################ 
  tauphi <- rgamma(1, c + (n-1) / 2, d + t(phi) %*% Q %*% phi / 2)
  
  #################################
  # Update Temporal CAR precision #
  ################################# 
  taupsi <- rgamma(1, e + (m-1) / 2, f + t(psi) %*% K %*% psi / 2)
  
  #####################################
  # Update interaction MCAR precision #
  ##################################### 
  tautheta <- rgamma(1, g + (n-1) * (m-1) / 2, h + sum(diag(t(thetam) %*% Q %*% thetam %*% K)) / 2)
  
  ###################################
  # Update augmented latent weights #
  ################################### 
  v <- c(rinvgauss(N, sqrt(xi^2 + 2*zeta) / abs(y - beta - Phic - Psic - Thetac),taue * (xi^2 + 2*zeta)))
  w <- 1/v
  
  ###################################
  # Update rate parameter for ALD   #
  ################################### 
  delta2 <- rgamma(1, a + 3*N/2, b + 1 / (2*zeta) * sum(v*(y - xi*w - beta - Phic - Psic - Thetac)^2) + sum(w))
  taue <- delta2 / zeta
  
  #################
  # Store Results #
  #################
  Beta[i] <- beta
  Phis[i,] <- phi
  Psis[i,] <- psi
  Thetas[i,] <- c(thetam)
  S2phi[i] <- 1 / tauphi
  S2psi[i] <- 1 / taupsi
  S2theta[i] <- 1 / tautheta
  Delta2[i] <- delta2 
  
  if (i%%100==0) print(i)
}

tottime <- proc.time() - ptm

###########
# Results #
###########
mbeta <- mean(Beta[-(1:burn)])
s2phi <- mean(S2phi[-(1:burn)])
s2psi <- mean(S2psi[-(1:burn)])
s2theta <- mean(S2theta[-(1:burn)])
delta2 <- mean(Delta2[-(1:burn)])

cat("Posterior Estimates", "\n", "beta = ", mbeta, "\n", "S2phi = ", s2phi,
    "\n", "S2psi = ", s2psi, "\n", "S2theta = ", s2theta, "\n", "Delat2 = ", delta2)   

#############
# Predict y #
#############
mphi <- colMeans(Phis[-(1:burn),])
mpsi <- colMeans(Psis[-(1:burn),])
mtheta <- colMeans(Thetas[-(1:burn),])

y.hat <- mbeta + rep(mphi, m) + rep(mpsi, each = n) + mtheta         # predicted outcomes

#####################################
# Agreement of Prediction and Truth #
#####################################
par(mfrow=c(1, 1),mar = c(5.1,5.1,2.1,2.1), oma = c(0,0,0,0))
plot(y.hat ~ unique(data$y50), cex = 1.5, xlab = expression(Q[50]), ylab = expression(hat(Q)), ylim = range(data$y50))
abline(a = 0, b = 1, lty = 2, lwd = 3, col = "darkgray")
